import os
from pathlib import Path
from signal import Signals
import subprocess
import sys

from . import builder
from . import github
from . import pypi
from .filesystem import try_delete, load_pyproject
from .log import logger


DOWNLOAD_SOURCES = {
    'gh': github,
    'github': github,
    'pypi': pypi.Downloader('pypi'),
    'testpypi': pypi.Downloader('testpypi'),
    'pypi-test': pypi.Downloader('testpypi'),  # for backwards compatibility
}


def aliases():
    """Returns a list of the aliases defined in pyproject.toml."""
    pyproject = load_pyproject()
    return pyproject.get('tool', {}).get('bork', {}).get('aliases', {})


def build():
    """Build the project."""
    try:
        builder.dist()
        builder.zipapp()

    except FileNotFoundError as e:
        if e.filename != 'pyproject.toml':
            raise e

        def setup(ext):
            return Path.cwd() / f"setup.{ext}"

        if setup("cfg").exists() or setup("py").exists():
            msg = """If you use setuptools, the following should be sufficient:

	[build-system]
	requires = ["setuptools > 42", "wheel"]
	build-backend = "setuptools.build_meta" """

        else:
            msg = "Please refer to your build system's documentation."

        logger().error(
            "You need a 'pyproject.toml' file describing which buildsystem to "
            "use, per PEP 517. %s", msg
        )

        raise e


def clean():
    """Removes artifacts generated by `bork clean`.

    (Specifically: `./build`, `./dist`, and any `*.egg-info` files.)
    """
    try_delete('./build')
    try_delete('./dist')
    for name in Path.cwd().glob('*.egg-info'):
        if name.is_dir():
            try_delete(name)


def download(package, release_tag, file_pattern, directory):
    """Saves files from the designated package, to the specified directory.

    The format for `package` is `<SOURCE>:<NAME>`, where:

    - `<SOURCE>` is one of: `gh`/`github`, `pypi`, or `testpypi`.
    - `<NAME>` is the identifier for the package on the specified source.

    E.g., Bork would be `gh:duckinator/bork` or `pypi:bork`.

    **NOTE:** `pypi-test` is a deprecated alias of `testpypi`, and will be
    removed in a future version.

    Arguments:
        package:
            Package to download files for, in the format `<SOURCE>:<NAME>`.

        release_tag:
            The version or tag to download.

        file_pattern:
            Any files matching this pattern will be downloaded.

            (Wildcards: `*` matches anything, `?` matches any single character.)

        directory:
            The directory where files are saved.
            This directory is created, if needed.
    """
    if file_pattern is None or len(file_pattern) == 0:
        raise ValueError('file_pattern must be non-empty.')

    if ':' not in package:
        raise ValueError('Invalid package/repository -- no source given.')

    source, package = package.split(':')

    if source not in DOWNLOAD_SOURCES:
        raise ValueError('Invalid package/repository -- unknown source given.')

    downloader = DOWNLOAD_SOURCES[source]
    downloader.download(package, release_tag, file_pattern, directory) # type:ignore


def release(repository_name, dry_run):
    """Uploads build artifacts to a PyPi instance or GitHub, as configured
    in pyproject.toml.

    Arguments:
        repository_name:
            The name of the PyPi repository. (You probably want 'pypi'.)

        dry_run:
            If True, don't actually release, just show what a release would do.
    """
    pyproject = load_pyproject()
    bork_config = pyproject.get('tool', {}).get('bork', {})
    release_config = bork_config.get('release', {})
    github_token = os.environ.get('BORK_GITHUB_TOKEN', None)
    version = builder.version_from_bdist_file()

    project_name = pyproject.get('project', {}).get('name', None)

    strip_zipapp_version = release_config.get('strip_zipapp_version', False)
    globs = release_config.get('github_release_globs', ['./dist/*.pyz'])

    release_to_github = release_config.get('github', False)
    release_to_pypi = release_config.get('pypi', True)

    if not release_to_github and not release_to_pypi:
        print('Configured to release to neither PyPi nor GitHub?')

    if release_to_github:
        github_repository = release_config.get('github_repository', None)

        if github_token is None:
            logger().error('No GitHub token specified. Use the BORK_GITHUB_TOKEN '
                'environment variable to set it.')

            if dry_run:
                logger().error('When not using --dry-run, this error is fatal.')
            else:
                sys.exit(1)

        config = github.GithubConfig(github_token, github_repository, project_name)
        github_release = github.GithubRelease(
            config, tag=f'v{version}', commitish=None, body=None,
            globs=globs,
            dry_run=dry_run, strip_zipapp_version=strip_zipapp_version)
        github_release.prepare()

    if release_to_pypi:
        pypi.upload(repository_name, './dist/*.tar.gz', './dist/*.whl',
                    dry_run=dry_run)

    if release_to_github:
        github_release.publish()


def run(alias):
    """Run the alias specified by `alias`, as defined in pyproject.toml."""
    pyproject = load_pyproject()

    try:
        commands = pyproject['tool']['bork']['aliases'][alias]
    except KeyError as error:
        raise RuntimeError(f"No such alias: '{alias}'") from error

    logger().info("Running '%s'", commands)

    if isinstance(commands, str):
        commands = [commands]
    elif isinstance(commands, list):
        pass
    else:
        raise TypeError(f"commands must be str or list, was {type(commands)}")

    try:
        for command in commands:
            print(command)
            subprocess.run(command, check=True, shell=True)

    except subprocess.CalledProcessError as error:
        if error.returncode < 0:
            signal = Signals(- error.returncode)
            msg = f"command '{error.cmd}' exited due to signal {signal.name} ({signal.value})"

        else:
            msg = f"bork: command '{error.cmd}' exited with error code {error.returncode}"

        raise RuntimeError(msg) from error



import os

import pytest
from helpers import bork_check, python_check, check_zipfile, check_tgz


@pytest.mark.network
def test_download(tmpdir):
    os.chdir(tmpdir)

    # Download a pyz file from GitHub, saved to ./downloads, and check
    # that it can be run with Python.
    bork_check("download", "gh:duckinator/emanate", "v7.0.0")
    python_check("downloads/emanate-7.0.0.pyz", "--help")

    # Download a pyz file for a specific version from GitHub, saved to ./bin
    bork_check("download", "gh:duckinator/emanate", "v7.0.0", "--directory", "bin/")
    python_check("bin/emanate-7.0.0.pyz", "--help")

    # Download a .tar.gz file from GitHub, saved to ./downloads
    bork_check("download",
               "gh:ppb/pursuedpybear", "v0.6.0",
               "--files", "*.tar.gz",
               "--directory", "downloads")
    assert check_tgz("downloads/ppb-0.6.0.tar.gz")

    # Download a .whl file from PyPi, and verify it's uncorrupted.
    bork_check("download", "pypi:emanate", "v6.0.0", "--files", "*.whl")
    assert check_zipfile("downloads/emanate-6.0.0-py3-none-any.whl")

    # Download a .whl file from PyPi's test instance, and verify it's uncorrupted.
    bork_check("download", "pypi-test:whaledo", "1.0.1", "--files", "*.whl")
    assert check_zipfile("downloads/whaledo-1.0.1-py3-none-any.whl")



from pathlib import Path
import shutil

import toml  # type: ignore


def load_pyproject():
    """
    Loads the pyproject.toml data.

    Will synthesize data if a legacy setuptools project is detected.
    """
    try:
        return toml.load('pyproject.toml')
    except FileNotFoundError:
        if Path('setup.py').exists() or Path('setup.cfg').exists():
            # Legacy project, use setuptools' legacy backend
            # (This backend is specified in PEP517)
            return {
                'build-system': {
                    'build-backend': 'setuptools.build_meta:__legacy__',
                    'requires': ['setuptools>=42'],
                },
            }
        # Can't figure out what kind of project it is.
        raise


def find_files(globs):
    files = []

    for glob in globs:
        matches = map(str, Path().glob(glob))
        files += list(matches)

    return files


def try_delete(path):
    if Path(path).is_dir():
        shutil.rmtree(path)
    elif Path(path).exists():
        raise RuntimeError(f"'{path}' is not a directory")



import json
from pathlib import Path
import subprocess
import urllib.request

import packaging.version

from .log import logger


class GithubApi:
    """
    Basic wrapper for the GitHub API.

    Usage:
        gh = GithubApi('duckinator', 'bork', '<token>')
        gh.create_release('TEST-RELEASE', assets={'dist/bork-4.0.5.pyz': 'bork.pyz'})
    """

    def __init__(self, owner, repo, project_name, token):
        self.owner = owner
        self.repo = repo
        self.token = token
        self.project_name = project_name
        self._last_release = None

    def publish(self, release):
        url = '/' + release['url'].split('/', 3)[3]
        return self._api_patch(url, {'draft': False})

    # pylint: disable=too-many-arguments,too-many-locals
    def create_release(self, tag_name, name=None, commitish=None, body=None, draft=True,
                       prerelease=None, assets=None):
        """
        `tag_name` is the name of the tag.
        `commitish` is a commit hash, branch, tag, etc.
        `body` is the body of the commit.
        `draft` indicates whether it should be a draft release or not.
        `prerelease` indicates whether it should be a prerelease or not.
        `assets` is a dict mapping local file paths to the uploaded asset name.
        """
        if commitish is None:
            commitish = self.run('git', 'rev-parse', 'HEAD')

        if body is None:
            body = '{repo} {tag}'

        if name is None:
            name = '{project_name} {tag}'

        if draft:
            draft_indicator = ' as a draft'
        else:
            draft_indicator = ''
        logger().info('Creating GitHub release %s%s. (commit=%s)', tag_name,
                      draft_indicator, commitish)

        if prerelease is None:
            prerelease = packaging.version.parse(tag_name).is_prerelease

        format_dict = {
            'project_name': self.project_name,
            'owner': self.owner,
            'repo': self.repo,
            'tag': tag_name,
            'tag_name': tag_name,
            'version': packaging.version.parse(tag_name).public,
        }

        # Don't fetch more data unless needed.
        if 'changelog' in body:
            format_dict['changelog'] = self.changelog()

        request = {
            'tag_name': tag_name,
            'target_commitish': commitish,
            'name': name.format(**format_dict),
            'body': body.format(**format_dict),
            'draft': draft,
            'prerelease': prerelease,
        }
        url = f"/repos/{self.owner}/{self.repo}/releases"
        response = self._api_post(url, request)

        upload_url = response['upload_url'].split('{?')[0]

        if assets:
            for local_file, asset_name in assets.items():
                self.add_release_asset(upload_url, local_file, asset_name)

        return response
    # pylint: enable=too-many-arguments,too-many-locals

    def changelog(self):
        prs = self._api_get(f'/repos/{self.owner}/{self.repo}/pulls?state=closed')
        prs = filter(self._relevant_to_changelog, prs)
        summaries = map(self._format_for_changelog, prs)
        return "\n".join(summaries)

    @staticmethod
    def _format_for_changelog(pr):
        return f'* {pr["title"]} (#{pr["number"]} by @{pr["user"]["login"]})'

    def _relevant_to_changelog(self, pr):
        if not pr or not pr['merged_at']:
            return False

        if pr['merged_at'] > self.last_release['created_at']:
            return True

        return False

    @property
    def last_release(self):
        if not self._last_release:
            self._last_release = self._api_get(
                f'/repos/{self.owner}/{self.repo}/releases')[0]
        return self._last_release

    def add_release_asset(self, upload_url, local_file, name):
        logger().info('Adding asset %s to release (original file: %s).',
                      name, local_file)

        data = Path(local_file).read_bytes()

        headers = {
            'Content-Type': 'application/octet-stream',
        }

        url = f"{upload_url}?name={name}"
        response = self._api_post(url, data, headers=headers, server='')
        return response

    @staticmethod
    def run(*command):
        return subprocess.check_output(command).decode().strip()

    # pylint: disable=too-many-arguments

    def _api_post(self, endpoint, data, headers=None, server=None, method=None):
        if method is None:
            method = 'POST'
        if headers is None:
            headers = {}
        if server is None:
            server = 'https://api.github.com'

        headers['Authorization'] = 'token ' + self.token
        headers['Accept'] = 'application/vnd.github.v3+json'

        if isinstance(data, (dict, list)):
            data = json.dumps(data).encode()

        req = urllib.request.Request(server + endpoint, data=data,
                                     headers=headers, method=method)
        logger().debug('%s %s', req.method, req.full_url)

        with urllib.request.urlopen(req) as f:
            response = f.read().decode()
        return json.loads(response)

    # pylint: enable=too-many-arguments

    def _api_get(self, endpoint, headers=None, server=None):
        return self._api_post(endpoint, None, headers, server, 'GET')

    def _api_patch(self, endpoint, data, headers=None, server=None):
        return self._api_post(endpoint, data, headers, server, 'PATCH')



import configparser
from pathlib import Path
import subprocess
import sys
# Slight kludge so we can have a function named zipapp().
import zipapp as Zipapp  # noqa: N812

import build

from .filesystem import load_pyproject, try_delete
# from .log import logger


# The "proper" way to handle the default would be to check python_requires
# in setup.cfg. But, since Bork needs Python 3, there's no point.
DEFAULT_PYTHON_INTERPRETER = '/usr/bin/env python3'


def dist():
    """Build the sdist and wheel distributions."""

    builder = build.ProjectBuilder('.')
    builder.build('sdist', './dist/')
    builder.build('wheel', './dist/')


def _setup_cfg_package_name():
    setup_cfg = configparser.ConfigParser()
    setup_cfg.read('setup.cfg')

    if 'metadata' not in setup_cfg or 'name' not in setup_cfg['metadata']:
        raise RuntimeError(
            "You need to set project.name in pyproject.toml OR metadata.name in setup.cfg"
        )

    return setup_cfg['metadata']['name']


def _python_interpreter(config):
    # To override the default interpreter, add this to your project's setup.cfg:
    #
    # [bork]
    # python_interpreter = /path/to/python
    return config.get('python_interpreter', DEFAULT_PYTHON_INTERPRETER)


def _bdist_file():
    return max(Path.cwd().glob('dist/*.whl'))


def _prepare_zipapp(dest, bdist_file):
    # Prepare zipapp directory.
    try_delete(dest)
    Path(dest).mkdir()
    return subprocess.check_call([
        sys.executable, '-m', 'pip', 'install', '--target', dest, bdist_file
    ])


def version_from_bdist_file():
    return _bdist_file().name.replace('.tar.gz', '').split('-')[1]


def zipapp():
    """
    Build a zipapp for the project.

    dist() should be called before zipapp().
    """

    pyproject = load_pyproject()
    config = pyproject.get('tool', {}).get('bork', {})
    zipapp_cfg = config.get('zipapp', {})
    want_zipapp = zipapp_cfg.get('enabled', False)

    if not want_zipapp:
        return

    # If the project name is specified in pyproject.toml, use it.
    # Otherwise, try getting it from setup.cfg.
    name = pyproject.get('project', {}).get('name', None)
    if name is None:
        name = _setup_cfg_package_name()
    dest = str(Path('build', 'zipapp'))
    version = version_from_bdist_file()
    main = zipapp_cfg['main']

    # Output file is dist/<package name>-<package version>.pyz.
    target = f"dist/{name}-{version}.pyz"

    _prepare_zipapp(dest, _bdist_file())

    Zipapp.create_archive(dest, target, _python_interpreter(config), main,
        compressed=True)
    if not Path(target).exists():
        raise RuntimeError(f"Failed to build zipapp: {target}")



import fnmatch
import json
from pathlib import Path
from urllib.request import urlopen
from typing import Optional

import packaging.version

from .asset_manager import download_assets
from .filesystem import find_files
from .github_api import GithubApi
from .log import logger


class GithubConfig:  # pylint: disable=too-few-public-methods
    def __init__(self, token: str, repository: str, project_name: str):
        owner, repo = repository.split('/')
        self.token = token
        self.owner = owner
        self.repo = repo

        if project_name is None:
            project_name = self.repo

        self.project_name = project_name


class GithubRelease:  # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-arguments
    def __init__(self, config: GithubConfig,
                 tag: str, commitish: Optional[str] = None,
                 body: Optional[str] = None, globs=None,
                 dry_run=False, prerelease=None, strip_zipapp_version=False):
        self.log = logger()

        self.owner = config.owner
        self.repo = config.repo
        self.token = config.token
        self.project_name = config.project_name

        self.tag_name = tag
        self.commitish = commitish
        self.body = body

        if self.body is None:
            self.body = self.release_template()

        self.prerelease = prerelease

        self.globs = globs
        if self.globs is None:
            self.globs = []

        self.dry_run = dry_run
        self.strip_zipapp_version = strip_zipapp_version

        self.github = None
        self.release = None

        self.assets = self._build_asset_list(find_files(globs))
    # pylint: enable=too-many-arguments

    def _build_asset_list(self, files):
        results = {}
        for file_path in files:
            filename = file_path.split('/')[-1]
            if self.strip_zipapp_version and filename.endswith('.pyz'):
                asset_name = '-'.join(filename.split('-')[:-1]) + '.pyz'
            else:
                asset_name = filename
            results[file_path] = asset_name
        return results

    def prepare(self):
        self.log.info('Preparing release for GitHub.')

        for (filename, asset_name) in self.assets.items():
            self.log.info('- %s as %s', filename, asset_name)

        if self.dry_run:
            self.log.warning(
                'Skipping creating draft GitHub release since this is a dry run.')
            return

        self.github = GithubApi(self.owner, self.repo, self.project_name, self.token)
        self.release = self.github.create_release(
            self.tag_name, commitish=self.commitish, body=self.body,
            draft=True, prerelease=self.prerelease,
            assets=self.assets)

    def publish(self):
        if self.dry_run:
            self.log.warning(
                'Skipping publishing GitHub release since this is a dry run.')
            return

        self.github.publish(self.release)

    @staticmethod
    def release_template():
        release_template_path = Path('.github/BORK_RELEASE_TEMPLATE.md')
        if release_template_path.exists():
            return release_template_path.read_text(encoding='utf-8')
        return None


def _relevant_asset(asset, file_pattern):
    file_patterns = file_pattern.split(',')
    for pattern in file_patterns:
        if fnmatch.fnmatch(asset['name'], pattern):
            return True
    return False


def _get_release_info(repo, name, draft=False, prerelease=False):
    if '/' not in repo:
        raise ValueError(
            f"repo must be of format <user>/<repo>, got '{repo}'",
        )

    log = logger()
    url = f"https://api.github.com/repos/{repo}/releases"
    with urlopen(url) as f:
        req = f.read().decode()
    releases = json.loads(req)

    try:
        if name == 'latest':
            # Filter out prereleases and drafts (unless specified in the arguments)
            releases = (
                r for r in releases
                if (draft or not r['draft'])
                and (prerelease or not r['prerelease'])  # noqa: W503
            )
            # Find the latest
            release = max(
                releases,
                key=lambda x: packaging.version.parse(x['tag_name']).public,
            )
            log.info("Selected release '%s' as latest", release['name'])
        else:
            release = list(filter(lambda x: x['tag_name'] == name, releases))[0]

    except (IndexError, ValueError) as e:
        raise RuntimeError(f"No such Github release: '{name}'") from e

    return release


def download(repo, release, file_pattern, directory):
    release_info = _get_release_info(repo, release, file_pattern)
    assets = filter(lambda x: _relevant_asset(x, file_pattern),
                    release_info['assets'])
    download_assets(assets, directory, url_key='browser_download_url')



"""Command-line interface for Bork.

Usage:

`bork COMMAND [OPTIONS] [ARGS]`

Options that exist for all commands include:

`--verbose`: enable verbose logging

`--debug`: enable even more verbose logging (sometimes too noisy to be helpful)

**Commands:**

"""

import argparse
import inspect
import logging
import os
import sys

from . import __version__
from . import api
from .log import logger


DOWNLOAD_SOURCES_STR = ' '.join(api.DOWNLOAD_SOURCES)


def aliases(_args):
    """
    ### `bork aliases`

    Prints a list of aliases (configured via pyproject.toml).
    """
    alias_list = api.aliases()
    if len(alias_list.keys()) == 0:
        print('No aliases available.')
        return

    for key in alias_list.keys():
        print(key)


def build(_args):
    """
    ### `bork build`

    Build the project.
    """
    api.build()


def clean(_args):
    """
    ### `bork clean`

    Remove files created by `bork build`.
    """
    api.clean()


def download(args):
    """
    ### `bork download [--files FILES] [--directory DIRECTORY] PACKAGE RELEASE`

    Download a release of the specified project.

    Arguments:
        --files=FILES:
            (default `*.pyz`)
            A comma-separated list of filenames to download.
            Supports wildcards (* = everything, ? = any single character).
        --directory=DIRECTORY:
            (default `downloads`)
            The directory to save files in. Created if missing.
        PACKAGE:
            The package to download. Of the format `SOURCE:PACKAGE_NAME`, where
            `PACKAGE_NAME` is the name of the package to download, and `SOURCE`
            is one of `gh`, `github`, `pypi`, or `testpypi`.
        RELEASE:
            The release or tag of the package that you want to download.
    """
    files = args.files
    directory = args.directory
    package = args.PACKAGE
    release_tag = args.RELEASE

    api.download(package, release_tag, files, directory)


def release(args):
    """
    ### `bork release [--pypi-repository=REPO | --test-pypi] [--dry-run]`

    Arguments:
        --pypi-repository=REPO:
            (default `pypi`)
            Repository to use. Valid values are pypi, testpypi, or anything
            defined in ".pypirc".

        --test-pypi:
            Equivalent to `--pypi-repository testpypi`

        --dry-run:
            Don't actually release, just show what a release would do.
    """
    pypi_repository = args.pypi_repository
    if args.test_pypi:
        pypi_repository = 'testpypi'
    api.release(pypi_repository, args.dry_run)


def run(args):
    """
    ### `bork run NAME`

    Run the alias specified by NAME, as defined in pyproject.toml.
    """
    api.run(args.ALIAS)


def _arg_parser():
    parser = argparse.ArgumentParser(
            prog="bork",
            description="A build and release tool for Python projects, with ZipApp support.")
    parser.add_argument("--version", action="store_true",
                        help="Print version information and exit.")
    parser.add_argument("--verbose", action="store_true",
                        help="Enable verbose logging.")
    parser.add_argument("--debug", action="store_true",
                        help="Enable VERY verbose logging. (Sometimes too noisy to be helpful.)")

    subparsers = parser.add_subparsers(title="Commands")
    aliasesp = subparsers.add_parser("aliases",
                                     help="Prints the aliases configured via pyproject.toml.")
    aliasesp.set_defaults(func=aliases)

    buildp = subparsers.add_parser("build", help="Build the project.")
    buildp.set_defaults(func=build)

    cleanp = subparsers.add_parser("clean", help="Remove files generated by `bork build`.")
    cleanp.set_defaults(func=clean)

    downloadp = subparsers.add_parser("download",
                                      help="Download a release of the specified project.")
    downloadp.add_argument("--files", action="store", default="*.pyz",
                          help="Comma-separated list of filenames to download."
                               "Supports wildcards (* = everything, ? = any single character).")
    downloadp.add_argument("--directory", action="store", default="downloads",
                           help="Directory to save files in. Created if missing. "
                                "(Default: `downloads`)")
    downloadp.add_argument("PACKAGE",
                           help="The package to download. Format: `SOURCE:PACKAGE_NAME`, where"
                                " `PACKAGE_NAME` is the name of the package to download, and "
                                "`SOURCE` is one of `gh`, `github`, `pypi`, or `testpypi`.")
    downloadp.add_argument("RELEASE", action="store", nargs='?', default="latest",
                           help="The release or tag to download.")
    downloadp.set_defaults(func=download)


    releasep = subparsers.add_parser("release", help="Publish a built project.")
    releasep.add_argument("--pypi-repository", default="pypi",
                         help="Repository to use. Valid values are pypi, "
                              "testpypi, or anything defined in '.pypirc'.")
    releasep.add_argument("--test-pypi", action="store_true",
                         help="Release to test.pypi.org instead of pypi.org.\n"
                              "Equivalent to '--pypi-repository testpypi'.")
    releasep.add_argument("--dry-run", action="store_true",
                         help="Don't actually release, just show what a release would do.")
    releasep.set_defaults(func=release)

    runp = subparsers.add_parser("run", help="Run the specified alias.")
    runp.add_argument("ALIAS")
    runp.set_defaults(func=run)

    return parser


def main(cmd_args=None):
    """
    Command-line entrypoint for bork.

    `cmd_args` should be either `None` or equivalent to `sys.argv[1:]`.
    """
    if sys.version_info < (3, 8):
        print('ERROR: Bork requires Python 3.8 or newer', file=sys.stderr)

    cmd_args = cmd_args or sys.argv[1:]
    if len(cmd_args) == 0:
        cmd_args = ["--help"]
    args = _arg_parser().parse_args(cmd_args)

    if args.version:
        print(f"bork v{__version__}")
        sys.exit()

    try:
        # pylint: disable=import-outside-toplevel
        import coloredlogs  # type: ignore
        # pylint: enable=import-outside-toplevel

        # Default to only printing WARNING and higher severity messages.
        log_level = logging.WARNING

        # If we got '--verbose', print INFO and higher severity messages.
        if args.verbose:
            log_level = logging.INFO

        # If we got '--debug', print DEBUG and higher severity messages.
        if args.debug:
            log_level = logging.DEBUG

        coloredlogs.install(
            level=log_level,
            fmt='%(name)s %(levelname)s %(message)s',
        )

    except ModuleNotFoundError:
        logging.basicConfig(
            level=logging.INFO if args.verbose else logging.WARNING,
            format='%(name)s %(levelname)s %(message)s',
        )

    try:
        args.func(args)
    except RuntimeError as err:
        thrower = inspect.trace()[-1]
        log = logger(thrower)

        (log.exception if args.verbose else log.error)(str(err))

        sys.exit(1)


def zipapp_main():
    # If Bork is put in a zipapp, this allows scripts executed as subprocesses
    # to access pep517.compat.
    #
    # The problem area is the `import compat` line in pep517's _in_process.py.
    # https://github.com/pypa/pep517/blob/master/pep517/_in_process.py
    os.environ['PYTHONPATH'] = ':'.join([*sys.path, sys.argv[0] + '/pep517'])

    main()



import os
from pathlib import Path

import pytest
from helpers import bork_check, check_run


def _repo_test(tmpdir, repo):
    os.chdir(tmpdir)
    check_run(['git', 'clone', '--depth', '1', repo, str(Path(tmpdir, 'repo'))])
    os.chdir('repo')
    bork_check('clean')
    bork_check('build')
    bork_check('release', '--dry-run')


@pytest.mark.network
@pytest.mark.slow
def test_repo_ppb_vector(tmpdir):
    _repo_test(tmpdir, 'https://github.com/ppb/ppb-vector.git')


@pytest.mark.network
@pytest.mark.slow
def test_repo_emanate(tmpdir):
    _repo_test(tmpdir, 'https://github.com/duckinator/emanate.git')


@pytest.mark.network
@pytest.mark.slow
def test_repo_bork(tmpdir):
    _repo_test(tmpdir, 'https://github.com/duckinator/bork.git')


@pytest.mark.network
@pytest.mark.slow
def test_repo_ppb_mutant(tmpdir):
    _repo_test(tmpdir, 'https://github.com/astronouth7303/ppb-mutant.git')
